/*
 *  xf86DuoPen 
 * 
 *  Based on the xf86Summa driver.
 *  
 *  Modified for the MGLOGIC DuoPen Tablet
 *   (c) 2001 Paul Ortyl <ortylp@from.pl>  
 * 
 */

/* $XFree86: xc/programs/Xserver/hw/xfree86/common/xf86DuoPen.c */
/* Driver version:    0.1.2
 * Tested with XFree86 v3.3.6
 * Modification date: 2001/03/12
 */

/* 
 * xf86Summa
 * 
 * Copyright 1996 by Steven Lang <tiger@tyger.org>
 * gpm summa-repeater support:
 * Sat Apr 22 15:01:47 CEST 1999 Frank Holtz <frank@hof-berlin.de>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Steven Lang not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Steven Lang makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STEVEN LANG DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL STEVEN LANG BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTIONS, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "Xos.h"
#include <signal.h>
#include <stdio.h>
#include <strings.h>

#define NEED_EVENTS
#include "X.h"
#include "Xproto.h"
#include "misc.h"
#include "inputstr.h"
#include "scrnintstr.h"
#include "XI.h"
#include "XIproto.h"

#if defined(sun) && !defined(i386)
#define POSIX_TTY
#include <errno.h>
#include <termio.h>
#include <fcntl.h>
#include <ctype.h>

#include "extio.h"
#else
#include "compiler.h"

#ifdef XFree86LOADER
#include "xf86_libc.h"
#endif
#include "xf86.h"
#include "xf86Procs.h"
#include "xf86_Config.h"
#include "xf86_OSlib.h"
#include "xf86Xinput.h"
#include "atKeynames.h"
#include "xf86Version.h"
#endif

#if !defined(sun) || defined(i386)
#include "osdep.h"
#include "exevents.h"

#include "extnsionst.h"
#include "extinit.h"
#endif

#if defined(__QNX__) || defined(__QNXNTO__)
#define POSIX_TTY
#endif
/*
** Debugging macros
*/
#ifdef DBG
#undef DBG
#endif
#ifdef DEBUG
#undef DEBUG
#endif


/*
** Contants and macro
*/
#define BUFFER_SIZE	256	/* size of reception buffer */
#define XI_NAME 	"duopad"	/* X device name for the stylus */

#define SYSCALL(call) while(((call) == -1) && (errno == EINTR))

#define SS_TABID0	"0"	/* Tablet ID 0 */
#define SS_FIRMID	"\005"	/* Request firmware ID string */

#define SS_PROMPT_MODE	"\006r"	/* Prompt mode */
#define SS_STREAM_MODE	"@"	/* Stream mode */

#define PHASING_BIT	0x80
#define TABID_BIT	0x20
#define XSIGN_BIT	0x10
#define YSIGN_BIT	0x08
#define BUTTON_BITS	0x07
#define COORD_BITS	0x7f

/* macro from counts/inch to counts/meter */
#define LPI2CPM(res)	(res * 1000 / 25.4)

/* config constants for "model" */
#define MODEL_AUTODETECT  1
#define MODEL_JX  2
#define MODEL_C   3

#ifndef XFREE86_V4

/*
** Configuration data
*/
#define DUOPEN_SECTION_NAME "DuoPen"
#define PORT		1
#define DEVICENAME	2
#define THE_MODE	3
#define CURSOR		4
#define INCREMENT	5
#define BORDER		6
#define DEBUG_LEVEL     7
#define HISTORY_SIZE	8
#define ALWAYS_CORE	9
#define ACTIVE_AREA	10
#define ACTIVE_OFFSET	11
#define COMPATIBLE	12
#define MODEL_TYPE   	13

#if !defined(sun) || defined(i386)
static SymTabRec penTab[] = {
	{ENDSUBSECTION,		"endsubsection"},
	{PORT,			"port"},
	{DEVICENAME,		"devicename"},
	{THE_MODE,		"mode"},
/*	{CURSOR,		"cursor"},*/
	{INCREMENT,		"increment"},
/*	{BORDER,		"border"},*/
	{DEBUG_LEVEL,		"debuglevel"},
/*	{HISTORY_SIZE,		"historysize"}, */
	{ALWAYS_CORE,		"alwayscore"},
	{ACTIVE_AREA,		"activearea"},
	{ACTIVE_OFFSET,		"activeoffset"},
/*	{COMPATIBLE,		"compatible"},*/
        {MODEL_TYPE,            "model"},
        {-1,			""}
};

#define RELATIVE	1
#define ABSOLUTE	2

static SymTabRec penModeTabRec[] = {
	{RELATIVE,	"relative"},
	{ABSOLUTE,	"absolute"},
	{-1,		""}
};

#define PUCK		1
#define STYLUS		2

/*
static SymTabRec penPointTabRec[] = {
	{PUCK,		"puck"},
	{STYLUS,	"stylus"},
	{-1,		""}
};
*/

static SymTabRec penModelTabRec[] = {
     {MODEL_AUTODETECT, "auto"},
     {MODEL_JX,	        "jx"  },
     {MODEL_C,          "c"   },
     {-1,	        ""    }};


#endif /* !defined(sun) || defined(i386) */

/*
** External declarations
*/
#if defined(sun) && !defined(i386)
#define ENQUEUE	suneqEnqueue
#else
#define ENQUEUE	xf86eqEnqueue

extern void xf86eqEnqueue(
#if NeedFunctionPrototypes
    xEventPtr /*e*/
#endif
);
#endif /* #if defined(sun) && !defined(i386) */

extern void miPointerDeltaCursor(
#if NeedFunctionPrototypes
    int /*dx*/,
    int /*dy*/,
    unsigned long /*time*/
#endif
);

#if !defined(sun) || defined(i386)
/*
** xf86PenConfig
** Reads the DuoPen section from the XF86Config file
*/
static Bool
xf86PenConfig(LocalDevicePtr *array, int inx, int max, LexPtr val)
{
    LocalDevicePtr	dev = array[inx];
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)(dev->private);
    int			token;
    int			mtoken;

    DBG(1, ErrorF("xf86PenConfig\n"));

    while ((token = xf86GetToken(penTab)) != ENDSUBSECTION) {
	switch(token) {
	case DEVICENAME:
	    if (xf86GetToken(NULL) != STRING)
		xf86ConfigError("Option string expected");
	    else {
		dev->name = strdup(val->str);
		if (xf86Verbose)
		    ErrorF("%s DuoPen X device name is %s\n", XCONFIG_GIVEN,
			   dev->name);
	    }
	    break;

	case PORT:
	    if (xf86GetToken(NULL) != STRING)
		xf86ConfigError("Option string expected");
	    else {
		priv->penDevice = strdup(val->str);
		if (xf86Verbose)
		    ErrorF("%s DuoPen port is %s\n", XCONFIG_GIVEN,
			   priv->penDevice);
	    }
	    break;

	 case MODEL_TYPE:
	   mtoken = xf86GetToken(penModelTabRec);
	   if ((mtoken == EOF) || (mtoken == STRING) || (mtoken == NUMBER)) 
	     xf86ConfigError("Model type token expected");
	   else {
	      switch (mtoken) {
	       case MODEL_AUTODETECT:
		 priv->model=MODEL_AUTODETECT;
		 /* just in case the autodetection fails
		  * select 6bit protocol
		  */
		 priv->flags &= ~EIGHTBIT_FLAG;
		 break;
	       case MODEL_JX:
		 priv->model=MODEL_JX;
		 priv->flags &= ~EIGHTBIT_FLAG;
		 break;
	       case MODEL_C:
		 priv->model=MODEL_C;
		 priv->flags |= EIGHTBIT_FLAG;
		 break;
	       default:
		 xf86ConfigError("Unknown Model type");
		 break;
	      }
	    }
	    break;


	case THE_MODE:
	    mtoken = xf86GetToken(penModeTabRec);
	    if ((mtoken == EOF) || (mtoken == STRING) || (mtoken == NUMBER)) 
		xf86ConfigError("Mode type token expected");
	    else {
		switch (mtoken) {
		case ABSOLUTE:
		    priv->flags |= ABSOLUTE_FLAG;
		    break;
		case RELATIVE:
		    priv->flags &= ~ABSOLUTE_FLAG;
		    break;
		default:
		    xf86ConfigError("Illegal Mode type");
		    break;
		}
	    }
	    break;

 /*
	case CURSOR:
	    mtoken = xf86GetToken(penPointTabRec);
	    if ((mtoken == EOF) || (mtoken == STRING) || (mtoken == NUMBER)) 
		xf86ConfigError("Cursor token expected");
	    else {
		switch (mtoken) {
		case STYLUS:
		    priv->flags |= STYLUS_FLAG;
		    break;
		case PUCK:
		    priv->flags &= ~STYLUS_FLAG;
		    break;
		default:
		    xf86ConfigError("Illegal cursor type");
		    break;
		}
	    }
	    break;
*/
/*	case INCREMENT:
	    if (xf86GetToken(NULL) != NUMBER)
		xf86ConfigError("Option number expected");
	    priv->penInc = val->num;
	    if (xf86Verbose)
		ErrorF("%s DuoPen increment value is %d\n", XCONFIG_GIVEN,
		       priv->penInc);
	    break;
*/
	case DEBUG_LEVEL:
	    if (xf86GetToken(NULL) != NUMBER)
		xf86ConfigError("Option number expected");
	    debug_level = val->num;
	    if (xf86Verbose) {
#if DEBUG
		ErrorF("%s DuoPen debug level sets to %d\n", XCONFIG_GIVEN,
		       debug_level);
#else
		ErrorF("%s DuoPen debug level not sets to %d because"
		       " debugging is not compiled\n", XCONFIG_GIVEN,
		       debug_level);
#endif
	    }
	    break;

/*	case HISTORY_SIZE:
	    if (xf86GetToken(NULL) != NUMBER)
		xf86ConfigError("Option number expected");
	    dev->history_size = val->num;
	    if (xf86Verbose)
		ErrorF("%s DuoPen Motion history size is %d\n", XCONFIG_GIVEN,
		       dev->history_size);      
	    break;
 */
/*
	case COMPATIBLE:
	    priv->flags |= COMPATIBLE_FLAG;
	    if(xf86Verbose)
		ErrorF("%s DuoPen compatible - will not query firmware ID\n", XCONFIG_GIVEN);
	    break;		    
*/
	case ALWAYS_CORE:
	    xf86AlwaysCore(dev, TRUE);
	    if (xf86Verbose)
		ErrorF("%s DuoPen device always stays core pointer\n",
		       XCONFIG_GIVEN);
	    break;

	case ACTIVE_AREA:
	   if (xf86GetToken(NULL) != NUMBER)
	     xf86ConfigError("Option number expected");
	   priv->penXSize = val->realnum * 100;
	   if (xf86GetToken(NULL) != NUMBER)
	     xf86ConfigError("Option number expected");
	   priv->penYSize = val->realnum * 100;
	   if (xf86Verbose)
	     ErrorF("%s DuoPen active area: %d.%02dx%d.%02d"
		    " inches\n", XCONFIG_GIVEN, priv->penXSize / 100,
		    priv->penXSize % 100, priv->penYSize / 100,
		    priv->penYSize % 100);
	   break;

	 case ACTIVE_OFFSET:
	   if (xf86GetToken(NULL) != NUMBER)
	     xf86ConfigError("Option number expected");
	   priv->penXOffset = val->realnum * 100;
	   if (xf86GetToken(NULL) != NUMBER)
	     xf86ConfigError("Option number expected");
	   priv->penYOffset = val->realnum * 100;
	   if (xf86Verbose)
	     ErrorF("%s DuoPen active area offsets: %d.%02d %d.%02d"
		    " inches\n", XCONFIG_GIVEN, priv->penXOffset / 100,
		    priv->penXOffset % 100, priv->penYOffset / 100,
		    priv->penYOffset % 100);
	   break;
	   
	 case EOF:
	    FatalError("Unexpected EOF (missing EndSubSection)");
	    break;

	default:
	    xf86ConfigError("DuoPen subsection keyword expected");
	    break;
	}
    }

    DBG(1, ErrorF("xf86PenConfig name=%s\n", priv->penDevice));

    return Success;
}
#endif /* !defined(sun) || defined(i386) */
#endif  /* #ifndef XFREE86_V4 */

/*
** xf86PenConvert
** Convert valuators to X and Y.
*/
static Bool
xf86PenConvert(LocalDevicePtr	local,
	       int		first,
	       int		num,
	       int		v0,
	       int		v1,
	       int		v2,
	       int		v3,
	       int		v4,
	       int		v5,
	       int*		x,
	       int*		y)
{
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;

    if (first != 0 || num == 1)
      return FALSE;
    *x = (v0 * screenInfo.screens[0]->width) / priv->penXSize;
    *y = (v1 * screenInfo.screens[0]->height) / priv->penYSize;
    if (*x < 0)
	*x = 0;
    if (*y < 0)
	*y = 0;
    if (*x > screenInfo.screens[0]->width)
	*x = screenInfo.screens[0]->width;
    if (*y > screenInfo.screens[0]->height)
	*y = screenInfo.screens[0]->height;

    DBG(6, ErrorF("Adjusted coords x=%d y=%d\n", *x, *y));

    return TRUE;
}

/*
** xf86PenReverseConvert
** Convert X and Y to valuators.
*/
static Bool
xf86PenReverseConvert(LocalDevicePtr	local,
		      int		x,
		      int		y,
		      int		*valuators)
{
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;
    valuators[0] = ((x * priv->penXSize) / screenInfo.screens[0]->width);
    valuators[1] = ((y * priv->penYSize) / screenInfo.screens[0]->height);

    DBG(6, ErrorF("Adjusted valuators v0=%d v1=%d\n", valuators[0], valuators[1]));

    return TRUE;
}

/*
** xf86PenReadInput
** Reads from the DuoPen and posts any new events to the server.
*/
static void
xf86PenReadInput(LocalDevicePtr local)
{
  DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;
  int		len, loop;
  int		is_absolute;
  int		x, y, buttons, prox, pressure;
  DeviceIntPtr	device;
  unsigned char	buffer[BUFFER_SIZE];
  
  DBG(7, ErrorF("xf86PenReadInput BEGIN device=%s fd=%d\n",
		priv->penDevice, local->fd));

  SYSCALL(len = read(local->fd, buffer, sizeof(buffer)));
  if (len <= 0) {
    ErrorF("error reading DuoPen device");
    return;
  }
  
  for(loop=0; loop<len; loop++) {
    if ((priv->penIndex == 0) && !(buffer[loop] & PHASING_BIT)) { 
      /* magic bit is not OK */
      DBG(6, ErrorF("xf86PenReadInput bad magic number 0x%x\n",
		    buffer[loop]));
      continue;
    }

    priv->penData[priv->penIndex++] = buffer[loop];
	
    if (priv->penIndex == (priv->flags & EIGHTBIT_FLAG ? 8: 6)) {
      /* the packet is OK */
      /* reset char count for next read */
      priv->penIndex = 0;
      
      if ((priv->flags)&EIGHTBIT_FLAG) {
	x=(((unsigned int) priv->penData[2])<<9)
	  |((unsigned int) priv->penData[1]<<2)
	  |((priv->penData[5]&0x60)>>5);
	y=(((unsigned int) priv->penData[4])<<9)
	  |((unsigned int)priv->penData[3]<<2)
	  |((priv->penData[5]&0x18)>>3);
	pressure=((priv->penData[7]&0x70)>>4)|((priv->penData[6]&0x3f)<<3);
	prox = (priv->penData[7] & 0x0F)? 0: 1;
      } else {
	/* additional test */
	if ((priv->penData[1] | priv->penData[2] | priv->penData[3] | 
	     priv->penData[4] | priv->penData[5]) & PHASING_BIT) {
	  DBG(6, ErrorF("xf86PenReadInput bad data in packet\n"));
	}
	 

	x=(((unsigned int) priv->penData[2])<<7)|priv->penData[1];
	y=(((unsigned int) priv->penData[4])<<7)|priv->penData[3];
	pressure=(((priv->penData[5])<<1)|
		  ((priv->penData[0]&0x20)>>5));
	prox = (priv->penData[0] & 0x40)? 0: 1;
      }
      y=priv->penMaxY - y;
      buttons = (priv->penData[0] & BUTTON_BITS);
      
      DBG(6, ErrorF("prox=%s\tx=%d\ty=%d\tbuttons=%d\toldx=%d\toldy=%d\n",
		    prox ? "true" : "false", x, y, buttons,
		    priv->penOldX,	priv->penOldY
		    ));
	  
      if (priv->flags & ABSOLUTE_FLAG) {
	x -= priv->penXOffset;
	y -= priv->penYOffset;
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x > priv->penXSize) x = priv->penXSize;
	if (y > priv->penYSize) y = priv->penYSize;
      } else {
	if ((!priv->penOldProximity) && (prox)) {
	  priv->penOldX=x;
	      priv->penOldY=y;
	      priv->penOldPressure=pressure;
	}
	x=x-priv->penOldX;
	y=y-priv->penOldY;
	pressure=pressure-priv->penOldPressure;
      }
	  
      device = local->dev;
      
      is_absolute = (priv->flags & ABSOLUTE_FLAG);
      
      /* coordinates are ready we can send events */
      if (prox) {
	if (!(priv->penOldProximity))
	  xf86PostProximityEvent(device, 1, 0, 3, x, y, pressure);
	
	if ((is_absolute && ((priv->penOldX != x) 
			     || (priv->penOldY != y)
			     || (priv->penOldPressure != pressure)
			     ))
	    || (!is_absolute && (x || y || pressure))) {
	  if (is_absolute || priv->penOldProximity) {
	    xf86PostMotionEvent(device, is_absolute, 0, 3, x, y,pressure);
	  }
	}
	{
	  int delta;
	  delta = buttons^(priv->penOldButtons);
	  if (delta) {
	    delta = buttons^(priv->penOldButtons);
	    /* test which button has been changed */
	    if (delta&0x01) {
	      xf86PostButtonEvent(device, is_absolute, 
				  1 , (buttons&0x01), 
				  0, 3, x, y,pressure);
	    };
	    if (delta&0x02) {
	      xf86PostButtonEvent(device, is_absolute, 
				  2 , (buttons&0x02), 
				  0, 3, x, y,pressure);
	    };
	    if (delta&0x04) {
	      xf86PostButtonEvent(device, is_absolute, 
				  3 , (buttons&0x04), 
				  0, 3, x, y,pressure);
	    };
	  }
	}
	priv->penOldButtons = buttons;
	if (is_absolute) {  
	  priv->penOldX = x;
	  priv->penOldY = y;
	  priv->penOldPressure = pressure;
	} else {
	  priv->penOldX += x;
	  priv->penOldY += y;
	  priv->penOldPressure += pressure;
	}
	priv->penOldProximity = prox;
      } else { /* !PROXIMITY */
	/* Any changes in buttons are ignored when !proximity */
	if (priv->penOldProximity)
	  xf86PostProximityEvent(device, 0, 0, 3, x, y, pressure);
	priv->penOldProximity = 0;
      }
    }
  }
  DBG(7, ErrorF("xf86PenReadInput END   device=0x%x priv=0x%x\n",
		local->dev, priv));
}

/*
** xf86PenControlProc
** It really does do something.  Honest!
*/
static void
xf86PenControlProc(DeviceIntPtr	device, PtrCtrl *ctrl)
{
    DBG(2, ErrorF("xf86PenControlProc\n"));
}

/*
** xf86PenWriteAndRead
** Write data, and get the response.
*/
static char *
xf86PenWriteAndRead(int fd, char *data, char *buffer, int len, int cr_term)
{
    int err, numread = 0;
    fd_set readfds;
    struct timeval timeout;
   
    SYSCALL(err = write(fd, data, strlen(data)));
    if (err == -1) {
	ErrorF("DuoPen write");
	return NULL;
    }

    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);
    
    while (numread < len) {
	timeout.tv_sec = 0;
	timeout.tv_usec = 200000;

	SYSCALL(err = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout));
	if (err == -1) {
	    ErrorF("DuoPen select");
	    return NULL;
	}
	if (!err) {
	    ErrorF("Timeout while reading DuoPen tablet. No tablet connected ???\n");
	    return NULL;
	}

	SYSCALL(err = read(fd, buffer + numread++, 1));
	if (err == -1) {
	    ErrorF("DuoPen read");
	    return NULL;
	}
	if (!err) {
	    --numread;
	    break;
	}
	if (cr_term && buffer[numread - 1] == '\r') {
	    buffer[numread - 1] = 0;
	    break;
	}
    }
    buffer[numread] = 0;
    return buffer;
}

/*
** xf86PenWriteAndReadn
** Write data, and get the response of unknown limited length
*/
static char *
xf86PenWriteAndReadn(int fd, char *data, char *buffer, int len, int cr_term)
{
    int err, numread = 0;
    fd_set readfds;
    struct timeval timeout;

    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);

   
    SYSCALL(err = write(fd, data, strlen(data)));
    if (err == -1) {
	ErrorF("DuoPen write");
	return NULL;
    }

    while (numread < len) {
	timeout.tv_sec = 0;
	timeout.tv_usec = 20000;

	SYSCALL(err = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout));
	if (err == -1) {
	    ErrorF("DuoPen select");
	    return NULL;
	}

       if (!err)
	 {break;}
       
	SYSCALL(err = read(fd, buffer + numread++, 1));
	if (err == -1) {
	    ErrorF("DuoPen read");
	    return NULL;
	}
	if (!err) {
	    --numread;
	    break;
	}
    }
    buffer[numread] = 0;
   return buffer;
}


/*
** xf86PenOpen
** Open and initialize the tablet, as well as probe for any needed data.
*/
static Bool
xf86PenOpen(LocalDevicePtr local)
{
    struct termios	termios_tty;
    struct timeval	timeout;
    char		buffer[256];
    char		dbuffer[256];
    int			err, idx;
    double		res100;
    double		sratio, tratio;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)local->private;

    DBG(1, ErrorF("opening %s\n", priv->penDevice));

    SYSCALL(local->fd = open(priv->penDevice, O_RDWR|O_NDELAY, 0));
    if (local->fd == -1) {
	ErrorF(priv->penDevice);
	return !Success;
    }
    DBG(2, ErrorF("%s opened as fd %d\n", priv->penDevice, local->fd));

#ifdef POSIX_TTY
    err = tcgetattr(local->fd, &termios_tty);
    if (err == -1) {
	ErrorF("DuoPen tcgetattr");
	return !Success;
    }
   
   if (priv->model==MODEL_AUTODETECT)
     {
	/* read (and forget) the PnP data */
	
	termios_tty.c_iflag=0;
	termios_tty.c_oflag=0;
	termios_tty.c_cflag=(B1200|CS7|CREAD|HUPCL);
	termios_tty.c_lflag=0;
	cfsetispeed(&termios_tty,B1200);
	cfsetospeed(&termios_tty,B1200);
	tcsetattr(local->fd,TCSANOW,&termios_tty);
	
	/* wait for the tablet to initialize itself after 
	 * powering itself (on open() call)
	 * 
	 */
	usleep(1000000);
     	err=read(local->fd,buffer,99);
   
	/* end of PnP */
     }
   termios_tty.c_iflag = 0;
   termios_tty.c_cflag = (B9600|CS8|CREAD|HUPCL|PARENB);
   termios_tty.c_lflag = 0;

/* prevent tty "terminal processing" */

    termios_tty.c_cc[VINTR] = 0;
    termios_tty.c_cc[VQUIT] = 0;
    termios_tty.c_cc[VERASE] = 0;
#ifdef VWERASE
    termios_tty.c_cc[VWERASE] = 0;
#endif
#ifdef VREPRINT
    termios_tty.c_cc[VREPRINT] = 0;
#endif
    termios_tty.c_cc[VKILL] = 0;
    termios_tty.c_cc[VEOF] = 0;
    termios_tty.c_cc[VEOL] = 0;
#ifdef VEOL2
    termios_tty.c_cc[VEOL2] = 0;
#endif
    termios_tty.c_cc[VSUSP] = 0;
#ifdef VDISCARD
    termios_tty.c_cc[VDISCARD] = 0;
#endif
#ifdef VLNEXT
    termios_tty.c_cc[VLNEXT] = 0; 
#endif

    termios_tty.c_cc[VMIN] = 1 ;
    termios_tty.c_cc[VTIME] = 10 ;

    cfsetospeed(&termios_tty,B9600);
    cfsetispeed(&termios_tty,B9600);

    err = tcsetattr(local->fd, TCSANOW, &termios_tty);
    if (err == -1) {
	ErrorF("DuoPen tcsetattr TCSANOW");
	return !Success;
    }
#else
    Code for someone else to write to handle OSs without POSIX tty functions
#endif

   DBG(1, ErrorF("initializing DuoPen tablet\n"));

/* Send zero to the tablet */
    SYSCALL(err = write(local->fd, "\000", 1));
    if (err == -1) {
	ErrorF("DuoPen write");
	return !Success;
    }

   if (xf86Verbose) {
     switch(priv->model)
	{
	 case MODEL_JX:
	   ErrorF("%s DuoPen firmware ID : %s\n", 
		  XCONFIG_GIVEN, 
		  "MGLOGIC JX-0605");
	   break;
	 case MODEL_C:
	   ErrorF("%s DuoPen firmware ID : %s\n", 
		  XCONFIG_GIVEN, 
		  "MGLOGIC C-Series");
	   break;
	 case MODEL_AUTODETECT:
	   ErrorF("%s DuoPen firmware -- autodetection forced\n", 
		  XCONFIG_GIVEN);
	   break;
	}
   }
   
   
   if (priv->model==MODEL_AUTODETECT)
     {
   
	/* Put it in prompt mode so it doesn't say anything before we're ready */
	SYSCALL(err = write(local->fd, SS_PROMPT_MODE, strlen(SS_PROMPT_MODE)));
	if (err == -1) {
	   ErrorF("DuoPen write");
	   return !Success;
	}
	
	/* wait 20 mSecs, just in case for the tablet to empty its buffers*/
	timeout.tv_sec = 0;
	timeout.tv_usec = 20000;
	SYSCALL(err = select(0, NULL, NULL, NULL, &timeout));
	if (err == -1) {
	   ErrorF("DuoPen select");
	   return !Success;
	}

	/* Clear any pending input */
	tcflush(local->fd, TCIFLUSH);
  
	if ( ! (priv->flags & COMPATIBLE_FLAG)) {
	   DBG(2, ErrorF("reading firmware ID\n"));
	   
	   if (!xf86PenWriteAndReadn(local->fd, SS_FIRMID, buffer, 200, 0))
	     return !Success;
	   
	   DBG(2, ErrorF("%s\n", buffer));
	   
	   if (!strncmp("MGLOGIC C-Series",buffer,16))
	     { priv->model=MODEL_C;}
	   else 
	   if (!strncmp("MGLOGIC JX-0605",buffer,15))
	     { priv->model=MODEL_JX;}
	   else  { 
	      ErrorF("%s DuoPen firmware -- unknown model, assuming JX-0605\n", XCONFIG_PROBED);
	      priv->model=MODEL_JX;
	   }
   
	   if (xf86Verbose)
	     ErrorF("%s DuoPen firmware ID : %s\n", XCONFIG_PROBED, buffer);
	}
     }
 
   switch(priv->model)
     {
      case MODEL_JX:
	/* MGLOGIC JX-0605 */
	priv->penRes=2560;  /* hardcoded resolution */   
	priv->penMaxX =15360;
	priv->penMaxY =12800;
	priv->penMaxPressure =255;

	break;
      case MODEL_C:
	/* MGLOGIC C-Series */
	priv->flags|=EIGHTBIT_FLAG;
	priv->penRes=2560;  /* hardcoded resolution */   
	priv->penMaxX =16200;
	priv->penMaxY =12200;
	priv->penMaxPressure =511;
	break;
	/* default: MGLOGIC_JX assumed
	 * 
	 */
     }
   res100=priv->penRes/100;
   
    if (xf86Verbose)
	ErrorF("%s DuoPen max tablet size %d.%02dinx%d.%02din, %dx%d "
	       "lines of resolution\n", XCONFIG_PROBED, 
	       priv->penMaxX / priv->penRes,
	       (priv->penMaxX * 100 / priv->penRes ) % 100,
	       priv->penMaxY / priv->penRes,
	       (priv->penMaxY * 100 / priv->penRes) % 100,
	       priv->penMaxX, priv->penMaxY);

    if (priv->penXOffset >= 0 && priv->penYOffset >= 0) {
	priv->penXOffset *= res100;
	priv->penYOffset *= res100;
	priv->penMaxX -= priv->penXOffset;
	priv->penMaxY -= priv->penYOffset;
    }

    if (priv->penXSize > 0 && priv->penYSize > 0) {
	if ((priv->penXSize * res100) <= priv->penMaxX &&
	    (priv->penYSize * res100) <= priv->penMaxY) {
	    priv->penXSize *= res100;
	    priv->penYSize *= res100;
	} else {
	    ErrorF("%s DuoPen active area bigger than tablet, "
		   "aspening maximum\n", XCONFIG_PROBED);
	    priv->penXSize = priv->penMaxX;
	    priv->penYSize = priv->penMaxY;
	}
    } else {
	priv->penXSize = priv->penMaxX;
	priv->penYSize = priv->penMaxY;
    }

    /* map tablet area by screen aspect ratio */
    sratio = (double)screenInfo.screens[0]->height /
	     (double)screenInfo.screens[0]->width;
    tratio = (double)priv->penMaxY / (double)priv->penMaxX;

    if (tratio <= 1.0) {		/* tablet horizontal > vertical */
	priv->penXSize = (double)priv->penYSize / sratio;
	if (priv->penXSize > priv->penMaxX) priv->penXSize = priv->penMaxX;
    }
    else {
	priv->penYSize = (double)priv->penXSize / sratio;
	if (priv->penYSize > priv->penMaxY) priv->penYSize = priv->penMaxY;
    }
    ErrorF("%s DuoPen using tablet area %d by %d, at res %d lpi\n",
	XCONFIG_PROBED, priv->penXSize, priv->penYSize, priv->penRes);

    if (priv->penInc > 95)
	priv->penInc = 95;
    if (priv->penInc < 0) {	/* increment value not given by user */
/* Make a guess as to the best increment value given video mode */
	if (priv->penXSize / screenInfo.screens[0]->width <
	       priv->penYSize / screenInfo.screens[0]->height)
	    priv->penInc = priv->penXSize / screenInfo.screens[0]->width;
	else
	    priv->penInc = priv->penYSize / screenInfo.screens[0]->height;
	if (priv->penInc < 1)
	    priv->penInc = 1;
	if (xf86Verbose)
	    ErrorF("%s Using increment value of %d\n", XCONFIG_PROBED,
		   priv->penInc);
    }

   /* Set the tablet to Stream Mode */
   
      SYSCALL(err = write(local->fd, SS_STREAM_MODE, strlen(SS_STREAM_MODE)));
      if (err == -1) {
	ErrorF("DuoPen write");
	return !Success;
      }

      if (err <= 0) {
	SYSCALL(close(local->fd));
	return !Success;
      }
    return Success;
}

/*
** xf86PenOpenDevice
** Opens and initializes the device driver stuff or penpthin.
*/
static int
xf86PenOpenDevice(DeviceIntPtr ppen)
{
    LocalDevicePtr	local = (LocalDevicePtr)ppen->public.devicePrivate;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)PRIVATE(ppen);

    if (xf86PenOpen(local) != Success) {
	if (local->fd >= 0) {
	    SYSCALL(close(local->fd));
	}
	local->fd = -1;
    }

/* Set the real values */
    InitValuatorAxisStruct(ppen,
			   0,
			   0, /* min val */
			   priv->penXSize, /* max val */
			   LPI2CPM(priv->penRes), /* resolution */
			   0, /* min_res */
			   LPI2CPM(priv->penRes)); /* max_res */
    InitValuatorAxisStruct(ppen,
			   1,
			   0, /* min val */
			   priv->penYSize, /* max val */
			   LPI2CPM(priv->penRes), /* resolution */
			   0, /* min_res */
			   LPI2CPM(priv->penRes)); /* max_res */
    InitValuatorAxisStruct(ppen,
			   2,
			   0, /* min val */
			   priv->penMaxPressure, /* max val */
			   1, /* resolution */
			   0, /* min_res */
			   1); /* max_res */

   return (local->fd != -1);
}

/*
** xf86PenProc
** Handle requests to do stuff to the driver.
*/
static int
xf86PenProc(DeviceIntPtr ppen, int what)
{
    CARD8		map[25];
    int			nbaxes;
    int			nbbuttons;
    int			loop;
    LocalDevicePtr	local = (LocalDevicePtr)ppen->public.devicePrivate;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)PRIVATE(ppen);

    DBG(2, ErrorF("BEGIN xf86PenProc dev=0x%x priv=0x%x what=%d\n", ppen, priv, what));

    switch (what) {
	case DEVICE_INIT:
	    DBG(1, ErrorF("xf86PenProc ppen=0x%x what=INIT\n", ppen));

	    nbaxes = 3;			/* X, Y, Pressure */
	    nbbuttons = 3;

	    for(loop=1; loop<=nbbuttons; loop++) map[loop] = loop;

	    if (InitButtonClassDeviceStruct(ppen,
					    nbbuttons,
					    map) == FALSE) {
		ErrorF("unable to allocate Button class device\n");
		return !Success;
	    }

	    if (InitFocusClassDeviceStruct(ppen) == FALSE) {
		ErrorF("unable to init Focus class device\n");
		return !Success;
	    }

	    if (InitPtrFeedbackClassDeviceStruct(ppen,
		   xf86PenControlProc) == FALSE) {
		ErrorF("unable to init ptr feedback\n");
		return !Success;
	    }

	    if (InitProximityClassDeviceStruct(ppen) == FALSE) {
		ErrorF("unable to init proximity class device\n"); 
		return !Success;
	    }

	    if (InitValuatorClassDeviceStruct(ppen,
		   nbaxes,
		   xf86GetMotionEvents,
		   local->history_size,
		   (priv->flags & ABSOLUTE_FLAG)? Absolute: Relative)
		   == FALSE) {
		ErrorF("unable to allocate Valuator class device\n"); 
		return !Success;
	    }
/* allocate the motion history buffer if needed */
	    xf86MotionHistoryAllocate(local);

	    AssignTypeAndName(ppen, local->atom, local->name);
/* open the device to gather informations */
	    xf86PenOpenDevice(ppen);
	    break;

	case DEVICE_ON:
	    DBG(1, ErrorF("xf86PenProc ppen=0x%x what=ON\n", ppen));

	    if ((local->fd < 0) && (!xf86PenOpenDevice(ppen))) {
		return !Success;
	    }
	    SYSCALL(write(local->fd, "P", 1));
	    AddEnabledDevice(local->fd);
	    ppen->public.on = TRUE;
	    break;

	case DEVICE_OFF:
	    DBG(1, ErrorF("xf86PenProc  ppen=0x%x what=%s\n", ppen,
		   (what == DEVICE_CLOSE) ? "CLOSE" : "OFF"));
	    if (local->fd >= 0)
		RemoveEnabledDevice(local->fd);
	    ppen->public.on = FALSE;
	    break;

	case DEVICE_CLOSE:
	    DBG(1, ErrorF("xf86PenProc  ppen=0x%x what=%s\n", ppen,
		   (what == DEVICE_CLOSE) ? "CLOSE" : "OFF"));
	    SYSCALL(close(local->fd));
	    local->fd = -1;
	    break;

	default:
	    ErrorF("unsupported mode=%d\n", what);
	    return !Success;
	    break;
    }
    DBG(2, ErrorF("END   xf86PenProc Success what=%d dev=0x%x priv=0x%x\n",
	   what, ppen, priv));
    return Success;
}

/*
** xf86PenClose
** It...  Uh...  Closes the physical device?
*/
static void
xf86PenClose(LocalDevicePtr local)
{
    if (local->fd >= 0) {
	SYSCALL(close(local->fd));
    }
    local->fd = -1;
}

/*
** xf86PenChangeControl
** When I figure out what it does, it will do it.
*/
static int
xf86PenChangeControl(LocalDevicePtr local, xDeviceCtl *control)
{
    xDeviceResolutionCtl	*res;

    res = (xDeviceResolutionCtl *)control;
	
    if ((control->control != DEVICE_RESOLUTION) ||
	   (res->num_valuators < 1))
	return (BadMatch);

    return(Success);
}

/*
** xf86PenSwitchMode
** Switches the mode.  For now just absolute or relative, hopefully
** more on the way.
*/
static int
xf86PenSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
{
    LocalDevicePtr	local = (LocalDevicePtr)dev->public.devicePrivate;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)(local->private);
    char		newmode;

    DBG(3, ErrorF("xf86PenSwitchMode dev=0x%x mode=%d\n", dev, mode));

    switch(mode) {
	case Absolute:
	    priv->flags |= ABSOLUTE_FLAG;
            break;

	case Relative:
	    priv->flags &= ~ABSOLUTE_FLAG;
	    break;

	default:
	    DBG(1, ErrorF("xf86PenSwitchMode dev=0x%x invalid mode=%d\n",
		   dev, mode));
	    return BadMatch;
    }
    return Success;
}

/*
** xf86PenAllocate
** Allocates the device structures for the DuoPen.
*/
static LocalDevicePtr
xf86PenAllocate()
{
    LocalDevicePtr	local = (LocalDevicePtr)xalloc(sizeof(LocalDeviceRec));
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)xalloc(sizeof(DuoPenDeviceRec));
#if defined (sun) && !defined(i386)
    char		*dev_name = getenv("DUOPEN_DEV");
#endif

    local->name = XI_NAME;
    local->type_name = "DuoPen Tablet";
    local->flags = 0; /*XI86_NO_OPEN_ON_INIT;*/
#ifndef XFREE86_V4
#if !defined(sun) || defined(i386)
    local->device_config = xf86PenConfig;
#endif
#endif
    local->device_control = xf86PenProc;
    local->read_input = xf86PenReadInput;
    local->control_proc = xf86PenChangeControl;
    local->close_proc = xf86PenClose;
    local->switch_mode = xf86PenSwitchMode;
    local->conversion_proc = xf86PenConvert;
    local->reverse_conversion_proc = xf86PenReverseConvert;
    local->fd = -1;
    local->atom = 0;
    local->dev = NULL;
    local->private = priv;
    local->private_flags = 0;
    local->history_size  = 0;

#if defined(sun) && !defined(i386)
    if (def_name) {
	priv->penDevice = (char *)xalloc(strlen(dev_name) + 1);
	strcpy(priv->penDevice, device_name);
	ErrorF("xf86PenOpen port changed to '%s'\n", priv->penDevice);
    } else {
	priv->penDevice = "";
    }
#else
    priv->penDevice = "";         /* device file name */
#endif
    priv->penInc = -1;            /* re-transmit position on increment */
    priv->penOldX = 0;           /* previous X position */
    priv->penOldY = 0;           /* previous Y position */
    priv->penOldProximity = 0;    /* previous proximity */
    priv->penOldPressure = 0;    /* previous pressure */
    priv->penOldButtons = 0;      /* previous buttons state */
    priv->penMaxX = -1;           /* max X value */
    priv->penMaxY = -1;           /* max Y value */
    priv->penXSize = -1;	  /* active area X */
    priv->penXOffset = 0;	  /* active area X offset */
    priv->penYSize = -1;	  /* active area Y */
    priv->penYOffset = 0;	  /* active area Y offset */
    priv->flags = ABSOLUTE_FLAG;  /* various flags -- default abs format */
    priv->penIndex = 0;           /* number of bytes read */
    priv->penRes = 0;		  /* resolution */
    priv->model=MODEL_AUTODETECT;

    return local;
}

#ifndef XFREE86_V4 

/*
** DuoPen device association
** Device section name and allocation function.
*/
DeviceAssocRec DuoPen_assoc =
{
  DUOPEN_SECTION_NAME,           /* config_section_name */
  xf86PenAllocate               /* device_allocate */
};

#ifdef DYNAMIC_MODULE
/*
** init_module
** Entry point for dynamic module.
*/
int
#ifndef DLSYM_BUG
init_module(unsigned long server_version)
#else
init_xf86DuoPen(unsigned long server_version)
#endif
{
    xf86AddDeviceAssoc(&DuoPen_assoc);

    if (server_version != XF86_VERSION_CURRENT) {
	ErrorF("Warning: DuoPen module compiled for version %s\n",
	       XF86_VERSION);
	return 0;
    } else {
	return 1;
    }
}
#endif

#ifdef XFree86LOADER
/*
 * Entry point for the loader code
 */
XF86ModuleVersionInfo xf86DuoPenVersion = {
    "xf86DuoPad",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XF86_VERSION_CURRENT,
    0x00010000,
    {0,0,0,0}
};

void
xf86DuoPenModuleInit(data, magic)
    pointer *data;
    INT32 *magic;
{
    static int cnt = 0;

    switch (cnt) {
      case 0:
      *magic = MAGIC_VERSION;
      *data = &xf86DuoPenVersion;
      cnt++;
      break;
      
      case 1:
      *magic = MAGIC_ADD_XINPUT_DEVICE;
      *data = &DuoPen_assoc;
      cnt++;
      break;

      default:
      *magic = MAGIC_DONE;
      *data = NULL;
      break;
    } 
}

#endif

#else

/************************************************************************/
/************************ XFree 4 ***************************************/
/************************************************************************/

static const char *default_options[] =
{
        "BaudRate",     "9600",
        "DataBits",     "8",
        "StopBits",     "1",
        "Parity",       "Odd",
        "FlowControl",  "Xoff",
        "VTime",        "10",
        "VMin",         "7",
        NULL
};

static InputDriverPtr duoDrv;

/*
 * xf86DuoUninit --
 *
 * called when the driver is unloaded.
 */
static void
xf86DuoUninit(InputDriverPtr	drv,
	      LocalDevicePtr	local,
	      int flags)
{
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;
    
    DBG(1, ErrorF("xf86DuoUninit\n"));
    
    xf86DuoProc(local->dev, DEVICE_OFF);
    
    xfree (priv);
    xf86DeleteInput(local, 0);    
}


/*
 * xf86HypInit --
 *
 * called when the module subsection is found in XF86Config
 */
static InputInfoPtr
xf86DuoInit(InputDriverPtr	drv,
	    IDevPtr		dev,
	    int			flags)
{
    LocalDevicePtr	local = NULL;
    DuoPenDevicePtr	priv = NULL;
    char		*s;

    duoDrv = drv;

    local = xf86DuoAllocate();
    local->conf_idev = dev;

    xf86CollectInputOptions(local, default_options, NULL);   
    xf86OptionListReport( local->options );

    if (local)
	priv = (DuoPenDevicePtr) local->private;

    if (!local || !priv) {
	goto SetupProc_fail;
    }
    
    local->name = dev->identifier;
    
    /* Serial Device is mandatory */
    priv->penDevice = xf86FindOptionValue(local->options, "Device");

    if (!priv->hypDevice) {
	xf86Msg (X_ERROR, "%s: No Device specified.\n", dev->identifier);
	goto SetupProc_fail;
    }

    /* Process the common options. */
    xf86ProcessCommonOptions(local, local->options);

    /* Optional configuration */

    xf86Msg(X_CONFIG, "%s serial device is %s\n", dev->identifier,
	    priv->penDevice);

    debug_level = xf86SetIntOption(local->options, "DebugLevel", 0);
    if (debug_level > 0) {
	xf86Msg(X_CONFIG, "DuoPen: debug level set to %d\n", debug_level);
    }



    s = xf86FindOptionValue(local->options, "Mode");

    if (s && (xf86NameCmp(s, "absolute") == 0)) {
	priv->flags = priv->flags | ABSOLUTE_FLAG;
    }
    else if (s && (xf86NameCmp(s, "relative") == 0)) {
	priv->flags = priv->flags & ~ABSOLUTE_FLAG;
    }
    else if (s) {
	xf86Msg(X_ERROR, "%s: invalid Mode (should be absolute or relative). "
		"Using default.\n", dev->identifier);
    }
    xf86Msg(X_CONFIG, "%s is in %s mode\n", local->name,
	    (priv->flags & ABSOLUTE_FLAG) ? "absolute" : "relative");	    


    s = xf86FindOptionValue(local->options, "Cursor");

    if (s && (xf86NameCmp(s, "stylus") == 0)) {
	priv->flags = priv->flags | STYLUS_FLAG;
    }
    else if (s && (xf86NameCmp(s, "puck") == 0)) {
	priv->flags = priv->flags & ~STYLUS_FLAG;
    }
    else if (s) {
	xf86Msg(X_ERROR, "%s: invalid Cursor (should be stylus or puck). "
		"Using default.\n", dev->identifier);
    }
    xf86Msg(X_CONFIG, "%s is in cursor-mode %s\n", local->name,
	    (priv->flags & STYLUS_FLAG) ? "cursor" : "puck");	    

    priv->hypXSize = xf86SetIntOption(local->options, "XSize", 0);
    if (priv->hypXSize != 0) {
	    xf86Msg(X_CONFIG, "%s: XSize = %d\n", 
		    dev->identifier, priv->hypXSize);
    }

    priv->hypYSize = xf86SetIntOption(local->options, "YSize", 0);
    if (priv->hypYSize != 0) {
	    xf86Msg(X_CONFIG, "%s: YSize = %d\n", 
		    dev->identifier, priv->hypYSize);
    }

    priv->hypXOffset = xf86SetIntOption(local->options, "XOffset", 0);
    if (priv->hypXOffset != 0) {
	    xf86Msg(X_CONFIG, "%s: XOffset = %d\n", 
		    dev->identifier, priv->hypXOffset);
    }

    priv->hypYOffset = xf86SetIntOption(local->options, "YOffset", 0);
    if (priv->hypYOffset != 0) {
	    xf86Msg(X_CONFIG, "%s: YOffset = %d\n", 
		    dev->identifier, priv->hypYOffset);
    }

    if (xf86SetBoolOption(local->options, "InvX", FALSE)) {
	    priv->flags |= INVX_FLAG;
    	    xf86Msg(X_CONFIG, "%s: InvX\n", dev->identifier); 
    }

    if (xf86SetBoolOption(local->options, "InvY", FALSE)) {
	    priv->flags |= INVY_FLAG;
    	    xf86Msg(X_CONFIG, "%s: InvY\n", dev->identifier); 
    }

    {
	int val;
	val = xf86SetIntOption(local->options, "BaudRate", 0);
	switch (val) {
	case 19200:
	    priv->flags |= BAUD_19200_FLAG;
	    break;
	case 9600:
	    priv->flags &= ~BAUD_19200_FLAG;
	    break;
	default:
    	    xf86Msg(X_CONFIG, "%s: Illegal speed value (must be 9600 or 19200)\n", dev->identifier); 
	    break;
 	}
    }

    /* mark the device configured */
    local->flags |= XI86_POINTER_CAPABLE | XI86_CONFIGURED;

    /* return the LocalDevice */
    return (local);

  SetupProc_fail:
    if (priv)
	xfree(priv);
    if (local)
	xfree(local);
    return NULL;
}

#ifdef XFree86LOADER
static
#endif
InputDriverRec DuoPen = {
    1,				/* driver version */
    "DuoPen",			/* driver name */
    NULL,			/* identify */
    xf86DuoInit,		/* pre-init */
    xf86DuoUninit,		/* un-init */
    NULL,			/* module */
    0				/* ref count */
};


/*
 ***************************************************************************
 *
 * Dynamic loading functions
 *
 ***************************************************************************
 */
#ifdef XFree86LOADER
/*
 * xf86DuoUnplug --
 *
 * called when the module subsection is found in XF86Config
 */
static void
xf86DuoUnplug(pointer	p)
{
    DBG(1, ErrorF("xf86DuoUnplug\n"));
}

/*
 * xf86DuoPlug --
 *
 * called when the module subsection is found in XF86Config
 */
static pointer
xf86DuoPlug(pointer	module,
	    pointer	options,
	    int		*errmaj,
	    int		*errmin)
{
    DBG(1, ErrorF("xf86DuoPlug\n"));
	
    xf86AddInputDriver(&DUOPEN, module, 0);

    return module;
}

static XF86ModuleVersionInfo xf86DuoVersionRec =
{
    "DuoPen",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XF86_VERSION_CURRENT,
    1, 0, 0,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}		/* signature, to be patched into the file by */
				/* a tool */
};

XF86ModuleData DuoPenModuleData = {&xf86DuoVersionRec,
					xf86DuoPlug,
					xf86DuoUnplug};

#endif /* XFree86LOADER */

#endif /* ifndef XFREE86_V4 */

/* end of xf86DuoPen.c */
