
#include <xf86Version.h>

#if XF86_VERSION_CURRENT >= XF86_VERSION_NUMERIC(3,9,0,0,0)
#define XFREE86_V4 1
#define XFree86LOADER 1
#endif

#include "Xos.h"
#include <signal.h>
#include <stdio.h>
#include <strings.h>

#define NEED_EVENTS
#include "X.h"
#include "Xproto.h"
#include "misc.h"
#include "inputstr.h"
#include "scrnintstr.h"
#include "XI.h"
#include "XIproto.h"

#if defined(sun) && !defined(i386)
#define POSIX_TTY
//#include <errno.h>
//#include <termio.h>
//#include <fcntl.h>
//#include <ctype.h>

#include "extio.h"
#else
#include "compiler.h"

#ifdef XFree86LOADER
//#include "xf86_libc.h"
#endif
#include "xf86.h"

#include "Configint.h"
#include "xf86tokens.h"
#include "xf86Priv.h"
#define XCONFIG_PROBED "(==)"
#define XCONFIG_GIVEN "(**)"

#include "xf86_OSlib.h"
#include "xf86Xinput.h"
#include "atKeynames.h"
#endif

#if !defined(sun) || defined(i386)
#include "osdep.h"
#include "exevents.h"
#include "extnsionst.h"
#include "extinit.h"
#endif

#if defined(__QNX__) || defined(__QNXNTO__)
#define POSIX_TTY
#endif

/************************************************************************/
/************************ XFree 4 ***************************************/
/************************************************************************/

static void xf86PenReadInput(LocalDevicePtr);
static int xf86PenChangeControl(LocalDevicePtr, xDeviceCtl *);
static void xf86PenClose(LocalDevicePtr);
static int xf86PenSwitchMode(ClientPtr, DeviceIntPtr, int);
static Bool xf86PenConvert(LocalDevicePtr	local,
	       int		first,
	       int		num,
	       int		v0,
	       int		v1,
	       int		v2,
	       int		v3,
	       int		v4,
	       int		v5,
	       int*		x,
	       int*		y);
static Bool xf86PenReverseConvert(LocalDevicePtr	local,
		      int		x,
		      int		y,
		      int		*valuators);



static LocalDevicePtr xf86PenAllocate();
static int xf86PenProc(DeviceIntPtr, int);

static const char *default_options[] =
{
  "devicename", "DuoPen",
  "model", "auto",
  "port", "/dev/ttyS1",
  "mode", "relative",
  NULL  
};

static InputDriverPtr duoDrv;

/*
 * xf86DuoUninit --
 *
 * called when the driver is unloaded.
 */
static void
xf86DuoUninit(InputDriverPtr	drv,
	      LocalDevicePtr	local,
	      int flags)
{
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;
    
    DBG(1, ErrorF("xf86DuoUninit\n"));
    
    xf86PenProc(local->dev, DEVICE_OFF);
    xfree (priv);
    xf86DeleteInput(local, 0);    
}


/*
 * xf86DuoInit --
 *
 * called when the module subsection is found in XF86Config
 */
static InputInfoPtr
xf86DuoInit(InputDriverPtr	drv,
	    IDevPtr		dev,
	    int			flags)
{
    LocalDevicePtr	local = NULL;
    DuoPenDevicePtr	priv = NULL;
    char		*s;

    duoDrv = drv;

    local = xf86PenAllocate();
    local->conf_idev = dev;
   
    xf86CollectInputOptions(local, default_options, NULL);
    xf86OptionListReport( local->options );
   
    if (local)
	priv = (DuoPenDevicePtr) local->private;

    if (!local || !priv) {
	goto SetupProc_fail;
    }
    
    local->name = dev->identifier;
   
    /* Serial Device is mandatory */
    priv->penDevice = xf86FindOptionValue(local->options, "Device");

    if (!priv->penDevice) {
	xf86Msg (X_ERROR, "%s: No Device specified.\n", dev->identifier);
	goto SetupProc_fail;
    }
   
    /* Process the common options. */
    xf86ProcessCommonOptions(local, local->options);
   
    /* Optional configuration */

    xf86Msg(X_CONFIG, "%s serial device is %s\n", dev->identifier,
	    priv->penDevice);

    debug_level = xf86SetIntOption(local->options, "DebugLevel", 0);
    if (debug_level > 0) {
	xf86Msg(X_CONFIG, "DuoPen: debug level set to %d\n", debug_level);
    }

    s = xf86FindOptionValue(local->options, "Mode");

    if (s && (xf86NameCmp(s, "absolute") == 0)) {
	priv->flags = priv->flags | ABSOLUTE_FLAG;
    }
    else if (s && (xf86NameCmp(s, "relative") == 0)) {
	priv->flags = priv->flags & ~ABSOLUTE_FLAG;
    }
    else if (s) {
	xf86Msg(X_ERROR, "%s: invalid Mode (should be absolute or relative). "
		"Using default.\n", dev->identifier);
    }
    xf86Msg(X_CONFIG, "%s is in %s mode\n", local->name,
	    (priv->flags & ABSOLUTE_FLAG) ? "absolute" : "relative");	    

   /**********************************************/
   s = xf86FindOptionValue(local->options, "Model");

    if (s && (xf86NameCmp(s, "jx") == 0)) {
	priv->model = MODEL_JX;
    } else if (s && (xf86NameCmp(s, "c") == 0)) {
	priv->model = MODEL_C;
    } else if (s && (xf86NameCmp(s, "auto") == 0)) {
	priv->model = MODEL_AUTODETECT;
    }
    else if (s) {
	xf86Msg(X_ERROR, "%s: invalid Model (should be jx or c or auto). "
		"Using default.\n", dev->identifier);
    }
   /**********************************************/
   
   
    s = xf86FindOptionValue(local->options, "Cursor");

    if (s && (xf86NameCmp(s, "stylus") == 0)) {
	priv->flags = priv->flags | STYLUS_FLAG;
    }
    else if (s && (xf86NameCmp(s, "puck") == 0)) {
	priv->flags = priv->flags & ~STYLUS_FLAG;
    }
    else if (s) {
	xf86Msg(X_ERROR, "%s: invalid Cursor (should be stylus or puck). "
		"Using default.\n", dev->identifier);
    }
    xf86Msg(X_CONFIG, "%s is in cursor-mode %s\n", local->name,
	    (priv->flags & STYLUS_FLAG) ? "cursor" : "puck");	    

    priv->penXSize = xf86SetIntOption(local->options, "XSize", 0);
    if (priv->penXSize != 0) {
	    xf86Msg(X_CONFIG, "%s: XSize = %d\n", 
		    dev->identifier, priv->penXSize);
    }

    priv->penYSize = xf86SetIntOption(local->options, "YSize", 0);
    if (priv->penYSize != 0) {
	    xf86Msg(X_CONFIG, "%s: YSize = %d\n", 
		    dev->identifier, priv->penYSize);
    }

    priv->penXOffset = xf86SetIntOption(local->options, "XOffset", 0);
    if (priv->penXOffset != 0) {
	    xf86Msg(X_CONFIG, "%s: XOffset = %d\n", 
		    dev->identifier, priv->penXOffset);
    }

    priv->penYOffset = xf86SetIntOption(local->options, "YOffset", 0);
    if (priv->penYOffset != 0) {
	    xf86Msg(X_CONFIG, "%s: YOffset = %d\n", 
		    dev->identifier, priv->penYOffset);
    }

    /* mark the device configured */
    local->flags |= XI86_POINTER_CAPABLE | XI86_CONFIGURED;
   
    /* return the LocalDevice */
    return (local);

  SetupProc_fail:
    if (priv)
	xfree(priv);
    if (local)
	xfree(local);
    return NULL;
}

#ifdef XFree86LOADER
static
#endif
InputDriverRec DuoPen = {
    1,				/* driver version */
    "DuoPen",			/* driver name */
    NULL,			/* identify */
    xf86DuoInit,		/* pre-init */
    xf86DuoUninit,		/* un-init */
    NULL,			/* module */
    0				/* ref count */
};


/*
 ***************************************************************************
 *
 * Dynamic loading functions
 *
 ***************************************************************************
 */
#ifdef XFree86LOADER
/*
 * xf86DuoUnplug --
 *
 * called when the module subsection is found in XF86Config
 */
static void
xf86DuoUnplug(pointer	p)
{
    DBG(1, ErrorF("xf86DuoUnplug\n"));
}

/*
 * xf86DuoPlug --
 *
 * called when the module subsection is found in XF86Config
 */
static pointer
xf86DuoPlug(pointer	module,
	    pointer	options,
	    int		*errmaj,
	    int		*errmin)
{
    DBG(1, ErrorF("xf86DuoPlug\n"));
	
    xf86AddInputDriver(&DuoPen, module, 0);

    return module;
}

static XF86ModuleVersionInfo xf86DuoVersionRec =
{
    "DuoPen",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XF86_VERSION_CURRENT,
    0, 1, 4,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}		/* signature, to be patched into the file by */
				/* a tool */
};

XF86ModuleData DuoPenModuleData = {&xf86DuoVersionRec,
					xf86DuoPlug,
					xf86DuoUnplug};

#endif /* XFree86LOADER */

/*
** xf86PenAllocate
** Allocates the device structures for the DuoPen.
*/
static LocalDevicePtr
xf86PenAllocate()
{
    LocalDevicePtr	local = xf86AllocateInput(duoDrv,0);
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)xalloc(sizeof(DuoPenDeviceRec));
#if defined (sun) && !defined(i386)
    char		*dev_name = getenv("DUOPEN_DEV");
#endif

    local->name = XI_NAME;
    local->type_name = "DuoPen Tablet";
    local->flags = 0; /*XI86_NO_OPEN_ON_INIT;*/
    local->device_control = xf86PenProc;
    local->read_input = xf86PenReadInput;
    local->control_proc = xf86PenChangeControl;
    local->close_proc = xf86PenClose;
    local->switch_mode = xf86PenSwitchMode;
    local->conversion_proc = xf86PenConvert;
    local->reverse_conversion_proc = xf86PenReverseConvert;
    local->fd = -1;
    local->atom = 0;
    local->dev = NULL;
    local->private = priv;
    local->private_flags = 0;
    local->history_size  = 0;

#if defined(sun) && !defined(i386)
    if (def_name) {
	priv->penDevice = (char *)xalloc(strlen(dev_name) + 1);
	strcpy(priv->penDevice, device_name);
	ErrorF("xf86PenOpen port changed to '%s'\n", priv->penDevice);
    } else {
	priv->penDevice = "";
    }
#else
    priv->penDevice = "";         /* device file name */
#endif
    priv->penInc = -1;            /* re-transmit position on increment */
    priv->penOldX = 0;           /* previous X position */
    priv->penOldY = 0;           /* previous Y position */
    priv->penOldProximity = 0;    /* previous proximity */
    priv->penOldPressure = 0;    /* previous pressure */
    priv->penOldButtons = 0;      /* previous buttons state */
    priv->penMaxX = -1;           /* max X value */
    priv->penMaxY = -1;           /* max Y value */
    priv->penXSize = -1;	  /* active area X */
    priv->penXOffset = 0;	  /* active area X offset */
    priv->penYSize = -1;	  /* active area Y */
    priv->penYOffset = 0;	  /* active area Y offset */
    priv->flags = ABSOLUTE_FLAG;  /* various flags -- default abs format */
    priv->penIndex = 0;           /* number of bytes read */
    priv->penRes = 0;		  /* resolution */
    priv->model=MODEL_AUTODETECT;

    return local;
}


/* end of xf86DuoPen.c */
