/*
 *  xf86DuoPen 
 * 
 *  Based on the xf86Summa driver.
 *  
 *  Modified for the MGLOGIC DuoPen Tablet
 *   (c) 2001 Paul Ortyl <ortylp@3miasto.net>
 * 
 */

/* TODO:
 * compile the driver under XFree-3.3.x
 * fix position & offset settings to be specified in cm or inches 
          rather than in pixels
 */

/* $XFree86: xc/programs/Xserver/hw/xfree86/common/xf86DuoPen.c */
/* Driver version:    0.1.5
 * Tested with XFree86 v4.[12].0
 * Modification date: 2002/03/23
 */

/* 
 * xf86Summa
 * 
 * Copyright 1996 by Steven Lang <tiger@tyger.org>
 * gpm summa-repeater support:
 * Sat Apr 22 15:01:47 CEST 1999 Frank Holtz <frank@hof-berlin.de>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Steven Lang not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Steven Lang makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STEVEN LANG DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL STEVEN LANG BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTIONS, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <xf86Version.h>

/*
** Device records
*/
#define ABSOLUTE_FLAG		1
#define STYLUS_FLAG		2
#define COMPATIBLE_FLAG		4
#define EIGHTBIT_FLAG		8


typedef struct 
{
    char	*penDevice;	/* device file name */
    int		penInc;		/* increment between transmits */
    int		penButTrans;	/* button translation flags */
    int		penOldX;	/* previous X position */
    int		penOldY;	/* previous Y position */
    int         penOldPressure;  /* previous pressure */
    int		penOldProximity; /* previous proximity */
    int		penOldButtons;	/* previous buttons state */
    int		penMaxX;	/* max X value */
    int		penMaxY;	/* max Y value */
    int         penMaxPressure;  /* max value of pressure */
    int		penXSize;	/* active area X size */
    int		penXOffset;	/* active area X offset */
    int		penYSize;	/* active area Y size */
    int		penYOffset;	/* active area Y offset */
    int		penRes;		/* resolution in lines per inch */
    int         model;          /* model type, 1 for autodetection */
    int		flags;		/* various flags */
    int		penIndex;	/* number of bytes read */
    unsigned char penData[8];	/* data read on the device */
} DuoPenDeviceRec, *DuoPenDevicePtr;


/*
** Contants and macro
*/
#define BUFFER_SIZE	256	/* size of reception buffer */
#define XI_NAME 	"duopad"	/* X device name for the stylus */

#define SYSCALL(call) while(((call) == -1) && (errno == EINTR))

#define SS_TABID0	"0"	/* Tablet ID 0 */
#define SS_FIRMID	"\005"	/* Request firmware ID string */

#define SS_PROMPT_MODE	"\006r"	/* Prompt mode */
#define SS_STREAM_MODE	"@"	/* Stream mode */

#define PHASING_BIT	0x80
#define TABID_BIT	0x20
#define XSIGN_BIT	0x10
#define YSIGN_BIT	0x08
#define BUTTON_BITS	0x07
#define COORD_BITS	0x7f

/* macro from counts/inch to counts/meter */
#define LPI2CPM(res)	(res * 1000 / 25.4)

/* config constants for "model" */
#define MODEL_AUTODETECT  1
#define MODEL_JX  2
#define MODEL_C   3

/*
** Device records
*/
#define ABSOLUTE_FLAG		1
#define STYLUS_FLAG		2
#define COMPATIBLE_FLAG		4
#define EIGHTBIT_FLAG		8


/* for debugging purposes */
static int      debug_level = 0;

#if XF86_VERSION_CURRENT >= XF86_VERSION_NUMERIC(3,9,0,0,0)
#define XFREE86_V4 1
#include "xf86DuoPen-X4.c"
#else
#include "xf86DuoPen-X3.c"
#endif


/*
** Debugging macros
*/
#ifdef DBG
#undef DBG
#endif
#ifdef DEBUG
#undef DEBUG
#endif


#define DEBUG	0
#if DEBUG
#define 	DBG(lvl, f) 	{if ((lvl) <= debug_level) f;}
#else
#define 	DBG(lvl, f)
#endif

/*
** xf86PenConvert
** Convert valuators to X and Y.
*/
static Bool
xf86PenConvert(LocalDevicePtr	local,
	       int		first,
	       int		num,
	       int		v0,
	       int		v1,
	       int		v2,
	       int		v3,
	       int		v4,
	       int		v5,
	       int*		x,
	       int*		y)
{
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;

    if (first != 0 || num == 1)
      return FALSE;

   *x = (v0 * screenInfo.screens[0]->width+priv->penXSize/2) 
     / priv->penXSize;
   *y = (v1 * screenInfo.screens[0]->height+priv->penYSize/2) 
     / priv->penYSize;

   
   if (*x < 0)
     *x = 0;
    if (*y < 0)
	*y = 0;
    if (*x > screenInfo.screens[0]->width)
	*x = screenInfo.screens[0]->width;
    if (*y > screenInfo.screens[0]->height)
	*y = screenInfo.screens[0]->height;

    DBG(6, ErrorF("Adjusted coords x=%d y=%d\n", *x, *y));

    return TRUE;
}

/*
** xf86PenReverseConvert
** Convert X and Y to valuators.
*/
static Bool
xf86PenReverseConvert(LocalDevicePtr	local,
		      int		x,
		      int		y,
		      int		*valuators)
{
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;
    valuators[0] = ((x * priv->penXSize+screenInfo.screens[0]->width/2) 
		    / screenInfo.screens[0]->width);
    valuators[1] = ((y * priv->penYSize+screenInfo.screens[0]->height/2) 
		    / screenInfo.screens[0]->height);

    DBG(6, ErrorF("Adjusted valuators v0=%d v1=%d\n", valuators[0], valuators[1]));

    return TRUE;
}

/*
** xf86PenReadInput
** Reads from the DuoPen and posts any new events to the server.
*/
static void
xf86PenReadInput(LocalDevicePtr local)
{
  DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;
  int		len, loop;
  int		is_absolute;
  int		x, y, buttons, prox, pressure;
  DeviceIntPtr	device;
  unsigned char	buffer[BUFFER_SIZE];
  
  DBG(7, ErrorF("xf86PenReadInput BEGIN device=%s fd=%d\n",
		priv->penDevice, local->fd));

  SYSCALL(len = read(local->fd, buffer, sizeof(buffer)));
  if (len <= 0) {
    ErrorF("error reading DuoPen device");
    return;
  }
  
  for(loop=0; loop<len; loop++) {
    if ((priv->penIndex == 0) && !(buffer[loop] & PHASING_BIT)) { 
      /* magic bit is not OK */
      DBG(6, ErrorF("xf86PenReadInput bad magic number 0x%x\n",
		    buffer[loop]));
      continue;
    }

    priv->penData[priv->penIndex++] = buffer[loop];
	
    if (priv->penIndex == (priv->flags & EIGHTBIT_FLAG ? 8: 6)) {
      /* the packet is OK */
      /* reset char count for next read */
      priv->penIndex = 0;
      
      if ((priv->flags)&EIGHTBIT_FLAG) {
	x=(((unsigned int) priv->penData[2])<<9)
	  |((unsigned int) priv->penData[1]<<2)
	  |((priv->penData[5]&0x60)>>5);
	y=(((unsigned int) priv->penData[4])<<9)
	  |((unsigned int)priv->penData[3]<<2)
	  |((priv->penData[5]&0x18)>>3);
	pressure=((priv->penData[7]&0x70)>>4)|((priv->penData[6]&0x3f)<<3);
	prox = (priv->penData[7] & 0x0F)? 0: 1;
      } else {
	/* additional test */
	if ((priv->penData[1] | priv->penData[2] | priv->penData[3] | 
	     priv->penData[4] | priv->penData[5]) & PHASING_BIT) {
	  DBG(6, ErrorF("xf86PenReadInput bad data in packet\n"));
	}
	 

	x=(((unsigned int) priv->penData[2])<<7)|priv->penData[1];
	y=(((unsigned int) priv->penData[4])<<7)|priv->penData[3];
	pressure=(((priv->penData[5])<<1)|
		  ((priv->penData[0]&0x20)>>5));
	prox = (priv->penData[0] & 0x40)? 0: 1;
      }
      y=priv->penMaxY - y;
      buttons = (priv->penData[0] & BUTTON_BITS);
      
      DBG(6, ErrorF("prox=%s\tx=%d\ty=%d\tbuttons=%d\toldx=%d\toldy=%d\n",
		    prox ? "true" : "false", x, y, buttons,
		    priv->penOldX,	priv->penOldY
		    ));
	  
      if (priv->flags & ABSOLUTE_FLAG) {
	x -= priv->penXOffset;
	y -= priv->penYOffset;
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x > priv->penXSize) x = priv->penXSize;
	if (y > priv->penYSize) y = priv->penYSize;
      } else {
	if ((!priv->penOldProximity) && (prox)) {
	  priv->penOldX=x;
	  priv->penOldY=y;
	  priv->penOldPressure=pressure;
	}
	x=x-priv->penOldX;
	y=y-priv->penOldY;
	pressure=pressure-priv->penOldPressure;
      }
	  
      device = local->dev;
      
      is_absolute = (priv->flags & ABSOLUTE_FLAG);
      
      /* coordinates are ready we can send events */
      if (prox) {
	if (!(priv->penOldProximity))
	  xf86PostProximityEvent(device, 1, 0, 3, x, y, pressure);
	
	if ((is_absolute && ((priv->penOldX != x) 
			     || (priv->penOldY != y)
			     || (priv->penOldPressure != pressure)
			     ))
	    || (!is_absolute && (x || y || pressure))) {
	  if (is_absolute || priv->penOldProximity) {
	    xf86PostMotionEvent(device, is_absolute, 0, 3, x, y,pressure);
	  }
	}
	{
	  int delta;
	  delta = buttons^(priv->penOldButtons);
	  if (delta) {
	    delta = buttons^(priv->penOldButtons);
	    /* test which button has been changed */
	    if (delta&0x01) {
	      xf86PostButtonEvent(device, is_absolute, 
				  1 , (buttons&0x01), 
				  0, 3, x, y,pressure);
	    };
	    if (delta&0x02) {
	      xf86PostButtonEvent(device, is_absolute, 
				  2 , (buttons&0x02), 
				  0, 3, x, y,pressure);
	    };
	    if (delta&0x04) {
	      xf86PostButtonEvent(device, is_absolute, 
				  3 , (buttons&0x04), 
				  0, 3, x, y,pressure);
	    };
	  }
	}
	priv->penOldButtons = buttons;
	if (is_absolute) {  
	  priv->penOldX = x;
	  priv->penOldY = y;
	  priv->penOldPressure = pressure;
	} else {
	  priv->penOldX += x;
	  priv->penOldY += y;
	  priv->penOldPressure += pressure;
	}
	priv->penOldProximity = prox;
      } else { /* !PROXIMITY */
	/* Any changes in buttons are ignored when !proximity */
	if (priv->penOldProximity)
	  xf86PostProximityEvent(device, 0, 0, 3, x, y, pressure);
	priv->penOldProximity = 0;
      }
    }
  }
  DBG(7, ErrorF("xf86PenReadInput END   device=0x%x priv=0x%x\n",
		local->dev, priv));
}

/*
** xf86PenControlProc
** It really does do something.  Honest!
*/
static void
xf86PenControlProc(DeviceIntPtr	device, PtrCtrl *ctrl)
{
    DBG(2, ErrorF("xf86PenControlProc\n"));
}

/*
** xf86PenWriteAndRead
** Write data, and get the response.
*/
static char *
xf86PenWriteAndRead(int fd, char *data, char *buffer, int len, int cr_term)
{
    int err, numread = 0;
    fd_set readfds;
    struct timeval timeout;
   
    SYSCALL(err = write(fd, data, strlen(data)));
    if (err == -1) {
	ErrorF("DuoPen write");
	return NULL;
    }

    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);
    
    while (numread < len) {
	timeout.tv_sec = 0;
	timeout.tv_usec = 200000;

	SYSCALL(err = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout));
	if (err == -1) {
	    ErrorF("DuoPen select");
	    return NULL;
	}
	if (!err) {
	    ErrorF("Timeout while reading DuoPen tablet. No tablet connected ???\n");
	    return NULL;
	}

	SYSCALL(err = read(fd, buffer + numread++, 1));
	if (err == -1) {
	    ErrorF("DuoPen read");
	    return NULL;
	}
	if (!err) {
	    --numread;
	    break;
	}
	if (cr_term && buffer[numread - 1] == '\r') {
	    buffer[numread - 1] = 0;
	    break;
	}
    }
    buffer[numread] = 0;
    return buffer;
}

/*
** xf86PenWriteAndReadn
** Write data, and get the response of unknown limited length
*/
static char *
xf86PenWriteAndReadn(int fd, char *data, char *buffer, int len, int cr_term)
{
    int err, numread = 0;
    fd_set readfds;
    struct timeval timeout;

    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);

   
    SYSCALL(err = write(fd, data, strlen(data)));
    if (err == -1) {
	ErrorF("DuoPen write");
	return NULL;
    }

    while (numread < len) {
	timeout.tv_sec = 0;
	timeout.tv_usec = 20000;

	SYSCALL(err = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout));
	if (err == -1) {
	    ErrorF("DuoPen select");
	    return NULL;
	}

       if (!err)
	 {break;}
       
	SYSCALL(err = read(fd, buffer + numread++, 1));
	if (err == -1) {
	    ErrorF("DuoPen read");
	    return NULL;
	}
	if (!err) {
	    --numread;
	    break;
	}
    }
    buffer[numread] = 0;
   return buffer;
}


/*
** xf86PenOpen
** Open and initialize the tablet, as well as probe for any needed data.
*/
static Bool
xf86PenOpen(LocalDevicePtr local)
{
    struct termios	termios_tty;
    struct timeval	timeout;
    char		buffer[256];
    char		dbuffer[256];
    int			err, idx;
    double		sratio, tratio;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)local->private;
   
    DBG(1, ErrorF("opening %s\n", priv->penDevice));

    SYSCALL(local->fd = open(priv->penDevice, O_RDWR|O_NDELAY, 0));
    if (local->fd == -1) {
	ErrorF(priv->penDevice);
	return !Success;
    }
    DBG(2, ErrorF("%s opened as fd %d\n", priv->penDevice, local->fd));

#ifdef POSIX_TTY
    err = tcgetattr(local->fd, &termios_tty);
    if (err == -1) {
	ErrorF("DuoPen tcgetattr");
	return !Success;
    }
   
   if (priv->model==MODEL_AUTODETECT)
     {
	/* read (and forget) the PnP data */
	
	termios_tty.c_iflag=0;
	termios_tty.c_oflag=0;
	termios_tty.c_cflag=(B1200|CS7|CREAD|HUPCL);
	termios_tty.c_lflag=0;
	cfsetispeed(&termios_tty,B1200);
	cfsetospeed(&termios_tty,B1200);
	tcsetattr(local->fd,TCSANOW,&termios_tty);
	
	/* wait for the tablet to initialize itself after 
	 * powering itself (on open() call)
	 * 
	 */
	usleep(1000000);
     	err=read(local->fd,buffer,99);
   
	/* end of PnP */
     }
   termios_tty.c_iflag = 0;
   termios_tty.c_cflag = (B9600|CS8|CREAD|HUPCL|PARENB);
   termios_tty.c_lflag = 0;

/* prevent tty "terminal processing" */

    termios_tty.c_cc[VINTR] = 0;
    termios_tty.c_cc[VQUIT] = 0;
    termios_tty.c_cc[VERASE] = 0;
#ifdef VWERASE
    termios_tty.c_cc[VWERASE] = 0;
#endif
#ifdef VREPRINT
    termios_tty.c_cc[VREPRINT] = 0;
#endif
    termios_tty.c_cc[VKILL] = 0;
    termios_tty.c_cc[VEOF] = 0;
    termios_tty.c_cc[VEOL] = 0;
#ifdef VEOL2
    termios_tty.c_cc[VEOL2] = 0;
#endif
    termios_tty.c_cc[VSUSP] = 0;
#ifdef VDISCARD
    termios_tty.c_cc[VDISCARD] = 0;
#endif
#ifdef VLNEXT
    termios_tty.c_cc[VLNEXT] = 0; 
#endif

    termios_tty.c_cc[VMIN] = 1 ;
    termios_tty.c_cc[VTIME] = 10 ;

    cfsetospeed(&termios_tty,B9600);
    cfsetispeed(&termios_tty,B9600);

    err = tcsetattr(local->fd, TCSANOW, &termios_tty);
    if (err == -1) {
	ErrorF("DuoPen tcsetattr TCSANOW");
	return !Success;
    }
#else
    Code for someone else to write to handle OSs without POSIX tty functions
#endif

   DBG(1, ErrorF("initializing DuoPen tablet\n"));

/* Send zero to the tablet */
    SYSCALL(err = write(local->fd, "\000", 1));
    if (err == -1) {
	ErrorF("DuoPen write");
	return !Success;
    }

   if (xf86Verbose) {
     switch(priv->model)
	{
	 case MODEL_JX:
	   ErrorF("%s DuoPen firmware ID : %s\n", 
		  XCONFIG_GIVEN, 
		  "MGLOGIC JX-0605");
	   break;
	 case MODEL_C:
	   ErrorF("%s DuoPen firmware ID : %s\n", 
		  XCONFIG_GIVEN, 
		  "MGLOGIC C-Series");
	   break;
	 case MODEL_AUTODETECT:
	   ErrorF("%s DuoPen firmware -- autodetection forced\n", 
		  XCONFIG_GIVEN);
	   break;
	}
   }
   
   
   if (priv->model==MODEL_AUTODETECT)
     {
   
	/* Put it in prompt mode so it doesn't say anything before we're ready */
	SYSCALL(err = write(local->fd, SS_PROMPT_MODE, strlen(SS_PROMPT_MODE)));
	if (err == -1) {
	   ErrorF("DuoPen write");
	   return !Success;
	}
	
	/* wait 20 mSecs, just in case for the tablet to empty its buffers*/
	timeout.tv_sec = 0;
	timeout.tv_usec = 20000;
	SYSCALL(err = select(0, NULL, NULL, NULL, &timeout));
	if (err == -1) {
	   ErrorF("DuoPen select");
	   return !Success;
	}

	/* Clear any pending input */
	tcflush(local->fd, TCIFLUSH);
  
	if ( ! (priv->flags & COMPATIBLE_FLAG)) {
	   DBG(2, ErrorF("reading firmware ID\n"));
	   
	   if (!xf86PenWriteAndReadn(local->fd, SS_FIRMID, buffer, 200, 0))
	     return !Success;
	   
	   DBG(2, ErrorF("%s\n", buffer));
	   
	   if (!strncmp("MGLOGIC C-Series",buffer,16))
	     { priv->model=MODEL_C;}
	   else 
	   if (!strncmp("MGLOGIC JX-0605",buffer,15))
	     { priv->model=MODEL_JX;}
	   else  { 
	      ErrorF("%s DuoPen firmware -- unknown model, assuming JX-0605\n", XCONFIG_PROBED);
	      priv->model=MODEL_JX;
	   }
   
	   if (xf86Verbose)
	     ErrorF("%s DuoPen firmware ID : %s\n", XCONFIG_PROBED, buffer);
	}
     }
 
   switch(priv->model)
     {
      case MODEL_JX:
	/* MGLOGIC JX-0605 */
	priv->penRes=2560;  /* hardcoded resolution */   
	priv->penMaxX =15360;
	priv->penMaxY =12800;
	priv->penMaxPressure =255;

	break;
      case MODEL_C:
	/* MGLOGIC C-Series */
	priv->flags|=EIGHTBIT_FLAG;
	priv->penRes=2560;  /* hardcoded resolution */   
	priv->penMaxX =16200;
	priv->penMaxY =12200;
	priv->penMaxPressure =511;
	break;
	/* default: MGLOGIC_JX assumed
	 * 
	 */
     }
   
    if (xf86Verbose)
	ErrorF("%s DuoPen max tablet size %d.%02dinx%d.%02din, %dx%d "
	       "lines of resolution\n", XCONFIG_PROBED, 
	       priv->penMaxX / priv->penRes,
	       (priv->penMaxX * 100 / priv->penRes ) % 100,
	       priv->penMaxY / priv->penRes,
	       (priv->penMaxY * 100 / priv->penRes) % 100,
	       priv->penMaxX, priv->penMaxY);

   if (priv->penXOffset >= 0 && priv->penYOffset >= 0
       && (priv->penXOffset<priv->penMaxX)
       && (priv->penYOffset<priv->penMaxY)
       )
     {
	priv->penMaxX -= priv->penXOffset;
	priv->penMaxY -= priv->penYOffset;
     };

    if (priv->penXSize > 0 && priv->penYSize > 0) {
	if ((priv->penXSize) <= priv->penMaxX &&
	    (priv->penYSize) <= priv->penMaxY) {
	    priv->penXSize = priv->penMaxX;
	    priv->penYSize = priv->penMaxY;
	} else {
	    ErrorF("%s DuoPen active area bigger than tablet, "
		   "aspening maximum\n", XCONFIG_PROBED);
	    priv->penXSize = priv->penMaxX;
	    priv->penYSize = priv->penMaxY;
	}
    } else {
	priv->penXSize = priv->penMaxX;
	priv->penYSize = priv->penMaxY;
    }

    /* map tablet area by screen aspect ratio */
    sratio = (double)screenInfo.screens[0]->height /
	     (double)screenInfo.screens[0]->width;
    tratio = (double)priv->penMaxY / (double)priv->penMaxX;

    if (tratio <= 1.0) {		/* tablet horizontal > vertical */
	priv->penXSize = (double)priv->penYSize / sratio;
	if (priv->penXSize > priv->penMaxX) priv->penXSize = priv->penMaxX;
    }
    else {
	priv->penYSize = (double)priv->penXSize / sratio;
	if (priv->penYSize > priv->penMaxY) priv->penYSize = priv->penMaxY;
    }
    ErrorF("%s DuoPen using tablet area %d by %d, at res %d lpi\n",
	XCONFIG_PROBED, priv->penXSize, priv->penYSize, priv->penRes);

    if (priv->penInc > 95)
	priv->penInc = 95;
    if (priv->penInc < 0) {	/* increment value not given by user */
/* Make a guess as to the best increment value given video mode */
	if (priv->penXSize / screenInfo.screens[0]->width <
	       priv->penYSize / screenInfo.screens[0]->height)
	    priv->penInc = priv->penXSize / screenInfo.screens[0]->width;
	else
	    priv->penInc = priv->penYSize / screenInfo.screens[0]->height;
	if (priv->penInc < 1)
	    priv->penInc = 1;
	if (xf86Verbose)
	    ErrorF("%s Using increment value of %d\n", XCONFIG_PROBED,
		   priv->penInc);
    }
   
   /* Set the tablet to Stream Mode */
   
      SYSCALL(err = write(local->fd, SS_STREAM_MODE, strlen(SS_STREAM_MODE)));
      if (err == -1) {
	ErrorF("DuoPen write");
	return !Success;
      }

      if (err <= 0) {
	SYSCALL(close(local->fd));
	return !Success;
      }
    return Success;
}

/*
** xf86PenOpenDevice
** Opens and initializes the device driver stuff or penpthin.
*/
static int
xf86PenOpenDevice(DeviceIntPtr ppen)
{
    LocalDevicePtr	local = (LocalDevicePtr)ppen->public.devicePrivate;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) local->private;

    if (xf86PenOpen(local) != Success) {
	if (local->fd >= 0) {
	    SYSCALL(close(local->fd));
	}
	local->fd = -1;
    }
   
/* Set the real values */
    InitValuatorAxisStruct(ppen,
			   0,
			   0, /* min val */
			   priv->penXSize, /* max val */
			   LPI2CPM(priv->penRes), /* resolution */
			   0, /* min_res */
			   LPI2CPM(priv->penRes)); /* max_res */
    InitValuatorAxisStruct(ppen,
			   1,
			   0, /* min val */
			   priv->penYSize, /* max val */
			   LPI2CPM(priv->penRes), /* resolution */
			   0, /* min_res */
			   LPI2CPM(priv->penRes)); /* max_res */
    InitValuatorAxisStruct(ppen,
			   2,
			   0, /* min val */
			   priv->penMaxPressure, /* max val */
			   priv->penMaxPressure+1, /* resolution */
			   0, /* min_res */
			   1); /* max_res */

   return (local->fd != -1);
}

/*
** xf86PenProc
** Handle requests to do stuff to the driver.
*/
static int
xf86PenProc(DeviceIntPtr ppen, int what)
{
    CARD8		map[25];
    int			nbaxes;
    int			nbbuttons;
    int			loop;
    LocalDevicePtr	local = (LocalDevicePtr) ppen->public.devicePrivate;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr) ppen;

    DBG(2, ErrorF("BEGIN xf86PenProc dev=0x%x priv=0x%x what=%d\n", ppen, priv, what));

    switch (what) {
	case DEVICE_INIT:
	    DBG(1, ErrorF("xf86PenProc ppen=0x%x what=INIT\n", ppen));

	    nbaxes = 3;			/* X, Y, Pressure */
	    nbbuttons = 3;

	    for(loop=1; loop<=nbbuttons; loop++) map[loop] = loop;

	    if (InitButtonClassDeviceStruct(ppen,
					    nbbuttons,
					    map) == FALSE) {
		ErrorF("unable to allocate Button class device\n");
		return !Success;
	    }

	    if (InitFocusClassDeviceStruct(ppen) == FALSE) {
		ErrorF("unable to init Focus class device\n");
		return !Success;
	    }

	    if (InitPtrFeedbackClassDeviceStruct(ppen,
		   xf86PenControlProc) == FALSE) {
		ErrorF("unable to init ptr feedback\n");
		return !Success;
	    }

	    if (InitProximityClassDeviceStruct(ppen) == FALSE) {
		ErrorF("unable to init proximity class device\n"); 
		return !Success;
	    }

	    if (InitValuatorClassDeviceStruct(ppen,
		   nbaxes,
		   xf86GetMotionEvents,
		   local->history_size,
		   (priv->flags & ABSOLUTE_FLAG)? Absolute: Relative)
		   == FALSE) {
		ErrorF("unable to allocate Valuator class device\n"); 
		return !Success;
	    }
/* allocate the motion history buffer if needed */
	    xf86MotionHistoryAllocate(local);

	    AssignTypeAndName(ppen, local->atom, local->name);
/* open the device to gather informations */
	    xf86PenOpenDevice(ppen);
	    break;

	case DEVICE_ON:
	    DBG(1, ErrorF("xf86PenProc ppen=0x%x what=ON\n", ppen));

	    if ((local->fd < 0) && (!xf86PenOpenDevice(ppen))) {
		return !Success;
	    }
	    SYSCALL(write(local->fd, "P", 1));
#ifdef XFREE86_V4
       xf86AddEnabledDevice(local);
#else
       AddEnabledDevice(local->fd);
#endif      
	    ppen->public.on = TRUE;
	    break;

	case DEVICE_OFF:
	    DBG(1, ErrorF("xf86PenProc  ppen=0x%x what=%s\n", ppen,
		   (what == DEVICE_CLOSE) ? "CLOSE" : "OFF"));
	    if (local->fd >= 0)
#ifdef XFREE86_V4
	 xf86RemoveEnabledDevice(local);
#else
         RemoveEnabledDevice(local->fd);
#endif      

       ppen->public.on = FALSE;
	    break;

	case DEVICE_CLOSE:
	    DBG(1, ErrorF("xf86PenProc  ppen=0x%x what=%s\n", ppen,
		   (what == DEVICE_CLOSE) ? "CLOSE" : "OFF"));
	    SYSCALL(close(local->fd));
	    local->fd = -1;
	    break;

	default:
	    ErrorF("unsupported mode=%d\n", what);
	    return !Success;
	    break;
    }
    DBG(2, ErrorF("END   xf86PenProc Success what=%d dev=0x%x priv=0x%x\n",
	   what, ppen, priv));
    return Success;
}

/*
** xf86PenClose
** It...  Uh...  Closes the physical device?
*/
static void
xf86PenClose(LocalDevicePtr local)
{
    if (local->fd >= 0) {
	SYSCALL(close(local->fd));
    }
    local->fd = -1;
}

/*
** xf86PenChangeControl
** When I figure out what it does, it will do it.
*/
static int
xf86PenChangeControl(LocalDevicePtr local, xDeviceCtl *control)
{
    xDeviceResolutionCtl	*res;

    res = (xDeviceResolutionCtl *)control;
	
    if ((control->control != DEVICE_RESOLUTION) ||
	   (res->num_valuators < 1))
	return (BadMatch);

    return(Success);
}

/*
** xf86PenSwitchMode
** Switches the mode.  For now just absolute or relative, hopefully
** more on the way.
*/
static int
xf86PenSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
{
    LocalDevicePtr	local = (LocalDevicePtr)dev->public.devicePrivate;
    DuoPenDevicePtr	priv = (DuoPenDevicePtr)(local->private);
    char		newmode;

    DBG(3, ErrorF("xf86PenSwitchMode dev=0x%x mode=%d\n", dev, mode));

    switch(mode) {
	case Absolute:
	    priv->flags |= ABSOLUTE_FLAG;
            break;

	case Relative:
	    priv->flags &= ~ABSOLUTE_FLAG;
	    break;

	default:
	    DBG(1, ErrorF("xf86PenSwitchMode dev=0x%x invalid mode=%d\n",
		   dev, mode));
	    return BadMatch;
    }
    return Success;
}

/* end of xf86DuoPen.c */
